from .utils import compare_versions
import functools
import json
import logging
import re

_log = logging.getLogger(__name__)


class SkipDrv(RuntimeError):
    """This derivation cannot be treated as package."""
    pass


# see parseDrvName built-in Nix function
# https://nixos.org/nix/manual/#ssec-builtins
R_VERSION = re.compile(r'^(\S+?)-([0-9]\S*)$')


def split_name(fullname):
    """Returns the pure package name and version of a derivation."""
    if fullname.endswith('.drv'):
        fullname = fullname[:-4]
    m = R_VERSION.match(fullname)
    if m:
        return m.group(1), m.group(2)
    return fullname, None


def load(path):
    with open(path) as f:
        d_obj = eval(f.read(), {'__builtins__': {}, 'Derive': Derive}, {})
    _log.debug('Loading drv %s', d_obj.name)
    d_obj.store_path = path
    return d_obj


def destructure(env):
    """Decodes Nix 2.0 __structuredAttrs."""
    return json.loads(env['__json'])


IGNORE_EXT = {
    '.tar.gz', '.tar.bz2', '.tar.xz', '.tar.lz', '.tgz', '.zip', '.gem',
    '.patch', '.patch.gz', '.patch.xz', '.diff',
}


@functools.total_ordering
class Derive(object):
    """Nix derivation as found as .drv files in the Nix store."""

    store_path = None

    def __init__(self, _outputs=None, _inputDrvs=None, _inputSrcs=None,
                 _system=None, _builder=None, _args=None,
                 envVars={}, _derivations=None, name=None, patches=None):
        """Create a derivation from a .drv file.

        The derivation files are just accidentally Python-syntax, but
        hey! :-)
        """
        envVars = dict(envVars)
        self.name = name or envVars.get('name')
        if not self.name:
            self.name = destructure(envVars)['name']
        for e in IGNORE_EXT:
            if self.name.endswith(e):
                raise SkipDrv()

        self.pname, self.version = split_name(self.name)
        if not self.version:
            raise SkipDrv()
        self.patches = patches or envVars.get('patches', '')

    def __repr__(self):
        return '<Derive({})>'.format(repr(self.name))

    def __eq__(self, other):
        if type(self) != type(other):
            return NotImplementedError()
        return self.name == other.name and self.patches == other.patches

    def __hash__(self):
        return hash(self.name)

    def __lt__(self, other):
        if self.pname < other.pname:
            return True
        if self.pname > other.pname:
            return False
        return compare_versions(self.version, other.version) == -1

    def __gt__(self, other):
        if self.pname > other.pname:
            return True
        if self.pname < other.pname:
            return False
        return compare_versions(self.version, other.version) == 1

    def product_candidates(self):
        """Return product name variations in order of preference."""
        underscore = self.pname.replace('-', '_')
        c = [self.pname, underscore, self.pname.lower(), underscore.lower()]
        yield c[0]
        for i in range(1, len(c)):
            if c[i] not in c[0:i]:
                yield c[i]

    def check(self, nvd):
        affected_by = set()
        patched_cves = self.applied_patches()
        for pname in self.product_candidates():
            for vuln in nvd.affected(pname, self.version):
                if vuln.cve_id not in patched_cves:
                    affected_by.add(vuln)
            if affected_by:
                # don't try further product candidates
                return affected_by
        return affected_by

    R_CVE = re.compile(r'CVE-\d{4}-\d+', flags=re.IGNORECASE)

    def applied_patches(self):
        """Guess which CVEs are patched from patch names."""
        return set(
            m.group(0).upper() for m in self.R_CVE.finditer(self.patches))



from vulnix.derivation import Derive
from vulnix.output import Filtered
from vulnix.vulnerability import Vulnerability
from vulnix.whitelist import Whitelist, WhitelistRule
import datetime
import freezegun
import io
import pytest

V = Vulnerability


@freezegun.freeze_time('2018-02-28')
def check_whitelist_entries(wl):
    assert len(wl) == 6

    entry = wl['*']
    assert entry.cve == {'CVE-2015-2504', 'CVE-2015-7696'}
    assert entry.issue_url == set()
    assert entry.comment == []
    assert not hasattr(entry, 'status')

    assert wl['libxslt'].comment == ["broken, won't fix"]
    assert wl['libxslt-2.0'].until == datetime.date(2018, 3, 1)
    assert wl['unzip'].cve == {'CVE-2015-7696'}
    assert wl['audiofile-0.3.2'] is not None

    entry = wl['audiofile-0.3.6']
    assert entry.issue_url == {'https://fb.flyingcircus.io/f/cases/26909/'}
    assert entry.cve == {'CVE-2017-6827', 'CVE-2017-6834', 'CVE-2017-6828'}


def test_from_yaml(whitelist_yaml):
    check_whitelist_entries(Whitelist.load(whitelist_yaml))


def test_from_toml(whitelist_toml):
    check_whitelist_entries(Whitelist.load(whitelist_toml))


def test_neither_name_nor_cve():
    with pytest.raises(RuntimeError):
        Whitelist.load(io.StringIO('-\n  comment: invalid entry\n'))


def test_parse_until():
    wl = Whitelist.load(io.StringIO('["libarchive"]\nuntil = "2019-03-10"\n'))
    assert len(wl.entries) == 1
    assert wl.entries['libarchive'].until == datetime.date(2019, 3, 10)


def test_match_pname_version():
    rule = WhitelistRule(pname='libxslt', version='2.0')
    assert rule.covers(Derive(name='libxslt-2.0'))
    assert not rule.covers(Derive(name='libxslt-2.1'))


def test_match_pname_only():
    rule = WhitelistRule(pname='libxslt', version='*')
    assert rule.covers(Derive(name='libxslt-2.0'))
    assert rule.covers(Derive(name='libxslt-2.1'))
    assert not rule.covers(Derive(name='libxml2-2.0'))


def test_match_pname_version_cve():
    rule = WhitelistRule(pname='cpio', version='2.12', cve=['CVE-2015-1197'])
    assert rule.covers(Derive(name='cpio-2.12'), {V('CVE-2015-1197')})
    assert not rule.covers(Derive(name='cpio-2.12'), {V('CVE-2015-1198')})


def test_match_cve_only():
    rule = WhitelistRule(cve=['CVE-2015-1197', 'CVE-2016-2037'])
    assert rule.covers(Derive(name='cpio-2.12'), {V('CVE-2015-1197')})
    assert not rule.covers(Derive(name='cpio-2.12'), {V('CVE-2016-2038')})


def test_match_partial():
    rule = WhitelistRule(cve=['CVE-2015-1197', 'CVE-2016-2037'])
    assert rule.covers(
            Derive(name='cpio-2.12'), {V('CVE-2015-1197'), V('CVE-2015-1198')})


def test_until(whitelist_toml):
    rule = WhitelistRule(pname='libxslt', until='2018-04-12')
    d = Derive(name='libxslt-2.0')
    with freezegun.freeze_time('2018-04-11'):
        assert rule.covers(d)
    with freezegun.freeze_time('2018-04-12'):
        assert not rule.covers(d)


def test_not_whitelisted(whitelist):
    filtered = whitelist.find(Derive(name='cpio-2.12'), {V('CVE-2016-2037')})
    assert filtered.rules == []
    assert filtered.report == {V('CVE-2016-2037')}


def test_filter(whitelist):
    affected = {
        # not filtered
        Derive(name='cpio-2.12'): {V('CVE-2016-2037')},
        # partially filtered
        Derive(name='audiofile-0.3.6'): {
            V('CVE-2017-6826'), V('CVE-2017-6827')},
        # fully filtered
        Derive(name='unzip-6.0'): {V('CVE-2015-7696')},
        # fully filtered w/o specific CVEs
        Derive(name='audiofile-0.3.2'): {V('CVE-2018-2668')},
    }
    f = whitelist.filter(affected)
    assert f[0].report == {V('CVE-2016-2037')}
    assert f[1].report == {V('CVE-2017-6826')}
    assert f[2].report == set()
    assert f[3].report == set()


def test_merge(whitelist):
    new = Whitelist.load(io.StringIO("""\
["libxslt-2.0"]
until = "2018-02-25"
comment = "latest date wins"

["audiofile-0.3.6"]
cve = ["CVE-2017-6827", "CVE-2017-6839"]
comment = "new stuff should be appended"
issue_url = "https://github.com/NixOS/nixpkgs/issues/30959"

["libtasn1-4.12"]
cve = ["CVE-2017-10790"]
"""))
    whitelist.merge(new)
    assert len(whitelist) == 7

    libxslt = whitelist['libxslt-2.0']
    assert libxslt.until == datetime.date(2018, 3, 1)
    assert libxslt.comment == ['latest date wins']

    audiofile = whitelist['audiofile-0.3.6']
    assert audiofile.cve == {
        'CVE-2017-6827',
        'CVE-2017-6834',
        'CVE-2017-6828',
        'CVE-2017-6839',
    }
    assert audiofile.comment == [
        'some issues not fixed upstream',
        'new stuff should be appended',
    ]
    assert audiofile.issue_url == {
        'https://fb.flyingcircus.io/f/cases/26909/',
        'https://github.com/NixOS/nixpkgs/issues/30959',
    }

    libtasn1 = whitelist['libtasn1-4.12']
    assert libtasn1.cve == {'CVE-2017-10790'}


def test_merge_into_empty():
    wl = Whitelist()
    new = Whitelist.load(io.StringIO("""\
["libxslt"]
["audiofile-0.3.6"]
"""))
    wl.merge(new)
    assert set(wl.entries.keys()) == {'libxslt', 'audiofile-0.3.6'}


def test_until_latest_wins(whitelist):
    new = Whitelist.load(io.StringIO("""\
["libxslt-2.0"]
until = "2018-03-02"

["audiofile-0.3.2"]
until = "2018-04-01"
"""))
    whitelist.merge(new)
    assert whitelist['libxslt-2.0'].until == datetime.date(2018, 3, 2)
    assert whitelist['audiofile-0.3.2'].until == datetime.date(2018, 4, 1)


@freezegun.freeze_time('2018-02-28')
def test_dump_str(whitelist):
    assert str(whitelist) == """\
["*"]
cve = [ "CVE-2015-2504", "CVE-2015-7696" ]

[libxslt]
comment = "broken, won't fix"

[unzip]
cve = "CVE-2015-7696"

["libxslt-2.0"]
cve = [ "CVE-2015-9019", "CVE-2017-2477" ]
until = "2018-03-01"

["audiofile-0.3.2"]

["audiofile-0.3.6"]
cve = [ "CVE-2017-6827", "CVE-2017-6828", "CVE-2017-6834" ]
comment = "some issues not fixed upstream"
issue_url = "https://fb.flyingcircus.io/f/cases/26909/"
"""


@freezegun.freeze_time('2018-03-01')
def test_dump_str_remove_outdated(whitelist):
    assert str(whitelist) == """\
["*"]
cve = [ "CVE-2015-2504", "CVE-2015-7696" ]

[libxslt]
comment = "broken, won't fix"

[unzip]
cve = "CVE-2015-7696"

["audiofile-0.3.2"]

["audiofile-0.3.6"]
cve = [ "CVE-2017-6827", "CVE-2017-6828", "CVE-2017-6834" ]
comment = "some issues not fixed upstream"
issue_url = "https://fb.flyingcircus.io/f/cases/26909/"
"""


@freezegun.freeze_time('2018-02-28')
def test_dump_add_cve(whitelist):
    whitelist.add_from(Filtered(Derive(name='libxslt-2.0'), {
        V('CVE-2019-13118'),
    }))
    assert """\
["libxslt-2.0"]
cve = [ "CVE-2015-9019", "CVE-2017-2477", "CVE-2019-13118" ]
until = "2018-03-01"
""" in str(whitelist)


def test_toml_missing_quote():
    t = io.StringIO("""\
[libxslt-2.0.1]
comment = "unquoted, triggers TOML's table syntax inadvertently"
""")
    with pytest.raises(RuntimeError):
        Whitelist.load(t)


def test_toml_malformed_url():
    with pytest.raises(ValueError):
        Whitelist.load(io.StringIO('["pkg"]\nissue_url = "foobar"'))


def test_section_header_unexpected_space():
    with pytest.raises(RuntimeError):
        Whitelist.load(io.StringIO("""
["ok-section-1.0"]

[ "broken-section-1.1" ]
comment = "whitespace confuses TOML parser"
"""))


def test_section_header_unexpected_space_2():
    with pytest.raises(RuntimeError):
        Whitelist.load(io.StringIO("""
["broken-section 1.2"]
comment = "incorrect whitespace between package and version"
"""))


def test_section_header_alphanumeric():
    Whitelist.load(io.StringIO("""
[systemd-236]
comment = "section headers consisting only of alphanum chars are ok"
"""))



from operator import attrgetter
import click
import functools
import json


def fmt_vuln(v):
    out = 'https://nvd.nist.gov/vuln/detail/{:17}'.format(v.cve_id)
    if v.cvssv3:
        out += ' {}'.format(v.cvssv3)
    return out.rstrip()


def vuln_sort_key(v):
    """Sort by CVSSv3 descending and CVE_ID ascending."""
    return (-v.cvssv3, v)


class Filtered:
    """Derivation with whitelist filtering applied.

    Initially, all CVEs are in the `report` set. When whitelist rules
    are added via `add()`, matching CVEs are moved into the `masked`
    set. Output formatting depends on which of these sets have any
    members.
    """

    until = None

    def __init__(self, derivation, vulnerabilities):
        self.derivation = derivation
        self.rules = []
        self.report = vulnerabilities
        self.masked = set()

    def __repr__(self):
        return '<Filtered({}, {}, {}, {})>'.format(
            self.derivation.pname, self.rules, len(self.report),
            len(self.masked))

    def add(self, wl_rule):
        self.rules.append(wl_rule)
        if wl_rule.until:
            if not self.until or self.until > wl_rule.until:
                self.until = wl_rule.until
        if wl_rule.cve:
            for r in wl_rule.cve:
                mask = set(vuln for vuln in self.report
                           if vuln.cve_id in wl_rule.cve)
                self.report -= mask
                self.masked |= mask
        else:
            self.masked |= self.report
            self.report = set()

    def print(self, show_masked=False):
        if not self.report and not show_masked:
            return
        d = self.derivation
        wl = not self.report

        click.secho('\n{}'.format('-' * 72), dim=wl)
        click.secho('{}\n'.format(d.name), fg='yellow', bold=True, dim=wl)
        if d.store_path:
            click.secho(d.store_path, fg='magenta', dim=wl)

        click.secho('{:50} {}'.format('CVE', 'CVSSv3'), dim=wl)
        for v in sorted(self.report, key=vuln_sort_key):
            click.echo(fmt_vuln(v))
        if show_masked:
            for v in sorted(self.masked, key=vuln_sort_key):
                click.secho("{}  [whitelisted]".format(fmt_vuln(v)), dim=True)

        issues = functools.reduce(
            set.union, (r.issue_url for r in self.rules), set())
        if issues:
            click.secho('\nIssue(s):', fg='cyan', dim=wl)
            for url in issues:
                click.secho(url, fg='cyan', dim=wl)
        for rule in self.rules:
            if rule.comment:
                click.secho('\nComment:', fg='blue', dim=wl)
                for comment in rule.comment:
                    click.secho('* ' + comment, fg='blue', dim=wl)


def output_text(vulns, show_whitelisted=False):
    report = [v for v in vulns if v.report]
    wl = [v for v in vulns if not v.report]

    if not report and not show_whitelisted:
        if wl:
            click.secho('Nothing to show, but {} left out due to whitelisting'.
                        format(len(wl)), fg='blue')
        else:
            click.secho('Found no advisories. Excellent!', fg='green')
        return

    click.secho('{} derivations with active advisories'.format(
        len(report)), fg='red')
    if wl and not show_whitelisted:
        click.secho('{} derivations left out due to whitelisting'.format(
            len(wl)), fg='blue')

    for i in sorted(report, key=attrgetter('derivation')):
        i.print(show_whitelisted)
    if show_whitelisted:
        for i in sorted(wl, key=attrgetter('derivation')):
            i.print(show_whitelisted)
    if wl and not show_whitelisted:
        click.secho('\nuse --show-whitelisted to see derivations with only '
                    'whitelisted CVEs', fg='blue')


def output_json(items, show_whitelisted=False):
    out = []
    for i in sorted(items, key=attrgetter('derivation')):
        if not i.report and not show_whitelisted:
            continue
        d = i.derivation
        out.append({
            'name': d.name,
            'pname': d.pname,
            'version': d.version,
            'derivation': d.store_path,
            'affected_by': sorted(v.cve_id for v in i.report),
            'whitelisted': sorted(v.cve_id for v in i.masked),
            'cvssv3_basescore':
                {v.cve_id: v.cvssv3 for v in (i.report | i.masked) if v.cvssv3}
        })
    print(json.dumps(out, indent=1))


def output(items, json=False, show_whitelisted=False):
    if json:
        output_json(items, show_whitelisted)
    else:
        output_text(items, show_whitelisted)
    if any(i.report for i in items):
        return 2
    if show_whitelisted and any(i.masked for i in items):
        return 1
    return 0



from http import HTTPStatus
from vulnix.nvd import NVD
from vulnix.whitelist import Whitelist
import hashlib
import http.server
import os
import os.path as p
import pkg_resources
import pytest
import threading


@pytest.fixture
def whitelist_toml():
    return pkg_resources.resource_stream(
        'vulnix', 'tests/fixtures/whitelist.toml')


@pytest.fixture
def whitelist_yaml():
    return pkg_resources.resource_stream(
        'vulnix', 'tests/fixtures/whitelist.yaml')


@pytest.fixture
def whitelist():
    return Whitelist.load(pkg_resources.resource_stream(
        'vulnix', 'tests/fixtures/whitelist.toml'))


class RequestHandler(http.server.SimpleHTTPRequestHandler):

    def do_GET(self):
        """Serve a GET request from the fixtures directory"""
        fn = p.join(p.dirname(__file__), 'fixtures', self.path[1:])
        print("path=", fn)
        try:
            with open(fn, 'rb') as f:
                stat = os.fstat(f.fileno())
                content = f.read()
        except OSError:
            self.send_error(HTTPStatus.NOT_FOUND)
        except IOError:
            self.send_error(HTTPStatus.INTERNAL_SERVER_ERROR)
        self.send_response(HTTPStatus.OK)
        self.send_header('Content-Type', self.guess_type(fn))
        self.send_header('Content-Length', stat.st_size)
        self.send_header('ETag', hashlib.sha1(content).hexdigest())
        self.end_headers()
        self.wfile.write(content)


@pytest.yield_fixture
def http_server():
    httpd = http.server.HTTPServer(("127.0.0.1", 0), RequestHandler)
    port = httpd.socket.getsockname()[1]
    mirror_url = 'http://127.0.0.1:{}/'.format(port)
    t = threading.Thread(target=httpd.serve_forever, daemon=True)
    t.start()
    yield mirror_url


@pytest.yield_fixture
def nvd(tmpdir, http_server):
    nvd = NVD(mirror=http_server, cache_dir=str(tmpdir))
    nvd.available_archives = ['modified']
    with nvd:
        yield nvd



from vulnix.utils import compare_versions, split_components, haskeys


def test_compare_versions():
    """Tests from https://nixos.org/nix/manual/#ssec-version-comparisons"""
    assert -1 == compare_versions('1.0', '2.3')
    assert -1 == compare_versions('2.1', '2.3')
    assert 0 == compare_versions('2.3', '2.3')
    assert 1 == compare_versions('2.5', '2.3')
    assert 1 == compare_versions('3.1', '2.3')
    assert 1 == compare_versions('2.3.1', '2.3')
    assert 1 == compare_versions('2.3.1', '2.3a')
    assert 1 == compare_versions('2.3', '2.3pre')
    assert -1 == compare_versions('2.3pre1', '2.3')
    assert -1 == compare_versions('2.3pre3', '2.3pre12')
    assert -1 == compare_versions('2.3a', '2.3c')
    assert -1 == compare_versions('2.3pre1', '2.3c')
    assert -1 == compare_versions('2.3pre1', '2.3q')


def test_split_components():
    assert ['2', '3', 'pre', '1'] == list(split_components('2.3pre1'))
    assert ['2019', '11', '01'] == list(split_components('2019-11-01'))
    assert ['5', '1', 'a', 'lts'] == list(split_components('5.1a-lts'))


def test_haskeys():
    assert not haskeys({}, 'foo')
    assert haskeys({'foo': 1}, 'foo')
    assert not haskeys({'foo': 1}, 'foo', 'bar')
    assert haskeys({'foo': {'bar': 1}}, 'foo', 'bar')
    assert not haskeys({'foo': {'bar': 1}}, 'foo', 'baz')



from vulnix.vulnerability import Vulnerability, Node
import json
import pkg_resources


def load(cve):
    return json.loads(pkg_resources.resource_string(
        'vulnix', 'tests/fixtures/{}.json'.format(cve)))


def test_update(nvd):
    nvd.update()
    assert len(nvd._root['advisory']) == 835
    cve = nvd.by_id('CVE-2010-0748')
    assert cve == Vulnerability.parse(load('CVE-2010-0748'))
    assert cve == nvd.by_product('transmission')[0]


def test_parse_vuln():
    v = Vulnerability.parse(load('CVE-2019-10160'))
    assert v.cve_id == 'CVE-2019-10160'
    assert v.nodes == [
        Node('python', 'python', '3.5'),
        Node('python', 'python', '3.6'),
        Node('python', 'python', '3.7'),
        Node('python', 'python', ['>=3.8.0a4', '<=3.8.0b1']),
    ]



from BTrees import OOBTree
from datetime import datetime, date, timedelta
from persistent import Persistent
from .vulnerability import Vulnerability
import fcntl
import glob
import gzip
import json
import logging
import os
import os.path as p
import requests
import transaction
import ZODB
import ZODB.FileStorage

DEFAULT_MIRROR = 'https://nvd.nist.gov/feeds/json/cve/1.1/'
DEFAULT_CACHE_DIR = '~/.cache/vulnix'

_log = logging.getLogger(__name__)


class NVD(object):
    """Access to the National Vulnerability Database.

    https://nvd.nist.gov/
    """

    def __init__(self, mirror=DEFAULT_MIRROR, cache_dir=DEFAULT_CACHE_DIR):
        self.mirror = mirror.rstrip('/') + '/'
        self.cache_dir = p.expanduser(cache_dir)
        current = date.today().year
        self.available_archives = [y for y in range(current-5, current+1)]

    def lock(self):
        self._lock = open(p.join(self.cache_dir, 'lock'), 'a')
        try:
            fcntl.lockf(self._lock, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except OSError:
            _log.info('Waiting for NVD lock...')
        fcntl.lockf(self._lock, fcntl.LOCK_EX)

    def __enter__(self):
        """Keeps database connection open while in this context."""
        _log.debug('Opening database in %s', self.cache_dir)
        os.makedirs(self.cache_dir, exist_ok=True)
        self.lock()
        self._db = ZODB.DB(ZODB.FileStorage.FileStorage(
            p.join(self.cache_dir, 'Data.fs')))
        self._connection = self._db.open()
        self._root = self._connection.root()
        try:
            self._root.setdefault('advisory', OOBTree.OOBTree())
            self._root.setdefault('by_product', OOBTree.OOBTree())
            self._root.setdefault('meta', Meta())
            # may trigger exceptions if the database is inconsistent
            list(self._root['by_product'].keys())
            if 'archives' in self._root:
                _log.warn('Pre-1.9.0 database found - rebuilding')
                self.reinit()
        except (TypeError, EOFError):
            _log.warn('Incompatible objects found in database - rebuilding DB')
            self.reinit()
        return self

    def __exit__(self, exc_type=None, exc_value=None, exc_tb=None):
        if exc_type is None:
            if self.meta.should_pack():
                _log.debug('Packing database')
                self._db.pack()
            transaction.commit()
        else:
            transaction.abort()
        self._connection.close()
        self._db.close()
        self._lock = None

    def reinit(self):
        """Remove old DB and rebuild it from scratch."""
        self._root = None
        transaction.abort()
        self._connection.close()
        self._db = None
        for f in glob.glob(p.join(self.cache_dir, "Data.fs*")):
            os.unlink(f)
        self._db = ZODB.DB(ZODB.FileStorage.FileStorage(
            p.join(self.cache_dir, 'Data.fs')))
        self._connection = self._db.open()
        self._root = self._connection.root()
        self._root['advisory'] = OOBTree.OOBTree()
        self._root['by_product'] = OOBTree.OOBTree()
        self._root['meta'] = Meta()

    @property
    def meta(self):
        return self._root['meta']

    def relevant_archives(self):
        """Returns list of NVD archives to check.

        If there was an update within the last two hours, nothing is
        done. If the last update was recent enough to be covered by
        the 'modified' feed, only that is checked. Else, all feeds
        are checked.
        """
        last_update = self.meta.last_update
        if last_update > datetime.now() - timedelta(hours=2):
            return []
        # the "modified" feed is sufficient if used frequently enough
        if last_update > datetime.now() - timedelta(days=7):
            return ['modified']
        return self.available_archives

    def update(self):
        """Download archives (if changed) and add CVEs to database."""
        changed = []
        for a in self.relevant_archives():
            arch = Archive(a)
            changed.append(arch.download(self.mirror, self.meta))
            self.add(arch)
        if any(changed):
            self.meta.last_update = datetime.now()
            self.reindex()

    def add(self, archive):
        advisories = self._root['advisory']
        for (cve_id, adv) in archive.items():
            advisories[cve_id] = adv

    def reindex(self):
        """Regenerate product index."""
        _log.info('Reindexing database')
        del self._root['by_product']
        bp = OOBTree.OOBTree()
        for vuln in self._root['advisory'].values():
            if vuln.nodes:
                for prod in (n.product for n in vuln.nodes):
                    bp.setdefault(prod, [])
                    bp[prod].append(vuln)
        self._root['by_product'] = bp
        transaction.commit()

    def by_id(self, cve_id):
        """Returns vuln or raises KeyError."""
        return self._root['advisory'][cve_id]

    def by_product(self, product):
        """Returns list of matching vulns or empty list."""
        try:
            return self._root['by_product'][product]
        except KeyError:
            return []

    def affected(self, pname, version):
        """Returns list of matching vulnerabilities."""
        res = set()
        for vuln in self.by_product(pname):
            if vuln.match(pname, version):
                res.add(vuln)
        return res


class Archive:

    """Single JSON data structure from NIST NVD."""

    def __init__(self, name):
        """Creates JSON feed object.

        `name` consists of a year or "modified".
        """
        self.name = name
        self.download_uri = 'nvdcve-1.1-{}.json.gz'.format(name)
        self.advisories = {}

    def download(self, mirror, meta):
        """Fetches compressed JSON data from NIST.

        Nothing is done if we have already seen the same version of
        the feed before.

        Returns True if anything has been loaded successfully.
        """
        url = mirror + self.download_uri
        _log.info('Loading %s', url)
        r = requests.get(url, headers=meta.headers_for(url))
        r.raise_for_status()
        if r.status_code == 200:
            _log.debug('Loading JSON feed "%s"', self.name)
            self.parse(gzip.decompress(r.content))
            meta.update_headers_for(url, r.headers)
            return True
        else:
            _log.debug('Skipping JSON feed "%s" (%s)', self.name, r.reason)
            return False

    def parse(self, nvd_json):
        added = 0
        raw = json.loads(nvd_json)
        for item in raw['CVE_Items']:
            try:
                vuln = Vulnerability.parse(item)
                self.advisories[vuln.cve_id] = vuln
                added += 1
            except ValueError:
                _log.debug('Failed to parse NVD item: %s', item)
        _log.debug("Added %s vulnerabilities", added)

    def items(self):
        return self.advisories.items()


class Meta(Persistent):
    """Metadate for database maintenance control"""

    pack_counter = 0
    last_update = datetime(1970, 1, 1)
    etag = None

    def should_pack(self):
        self.pack_counter += 1
        if self.pack_counter > 25:
            self.pack_counter = 0
            return True
        return False

    def headers_for(self, url):
        """Returns dict of additional request headers."""
        if self.etag and url in self.etag:
            return {'If-None-Match': self.etag[url]}
        return {}

    def update_headers_for(self, url, resp_headers):
        """Updates self from HTTP response headers."""
        if 'ETag' in resp_headers:
            if self.etag is None:
                self.etag = OOBTree.OOBTree()
            self.etag[url] = resp_headers['ETag']



from .derivation import load, SkipDrv, Derive
from .utils import call
import os.path as p
import json
import logging

_log = logging.getLogger(__name__)


class Store(object):

    def __init__(self, requisites=True):
        self.requisites = requisites
        self.derivations = set()
        self.experimental_flag_needed = None

    def add_gc_roots(self):
        """Add derivations found for all live GC roots.

        Note that this usually includes old system versions.
        """
        _log.debug('Loading all live derivations')
        for d in call(['nix-store', '--gc', '--print-live']).splitlines():
            self.update(d)

    def add_profile(self, profile):
        """Add derivations found in this nix profile."""
        json_manifest_path = p.join(profile, 'manifest.json')
        if p.exists(json_manifest_path):
            _log.debug('Loading derivations from {}'.format(
                json_manifest_path))
            with open(json_manifest_path, 'r') as f:
                json_manifest = json.load(f)
            if json_manifest['version'] > 1:
                raise RuntimeError(
                    ('Profile manifest.json version {} ' +
                     'not yet supported').format(
                        json_manifest['version']))
            for element in json_manifest['elements']:
                if element['active']:
                    for path in element['storePaths']:
                        self.add_path(path)
        else:
            _log.debug('Loading derivations from user profile {}'.format(
                profile))
            for line in call(['nix-env', '-q', '--out-path',
                              '--profile', profile]).splitlines():
                self.add_path(line.split()[1])

    def _call_nix(self, args):
        if self.experimental_flag_needed is None:
            self.experimental_flag_needed = (
                '--experimental-features' in call(['nix', '--help']))

        if self.experimental_flag_needed:
            return call(['nix',
                         '--experimental-features',
                         'nix-command'] + args)
        return call(['nix'] + args)

    def _find_deriver(self, path):
        if path.endswith('.drv'):
            return path
        # Deriver from QueryPathInfo
        qpi_deriver = call(['nix-store', '-qd', path]).strip()
        _log.debug('qpi_deriver: %s', qpi_deriver)
        if qpi_deriver and qpi_deriver != 'unknown-deriver' and p.exists(
                qpi_deriver):
            return qpi_deriver
        # Deriver from QueryValidDerivers
        qvd_deriver = list(json.loads(
            self._call_nix(['show-derivation', path])).keys())[0]
        _log.debug('qvd_deriver: %s', qvd_deriver)
        if qvd_deriver and p.exists(qvd_deriver):
            return qvd_deriver

        error = ""
        if qpi_deriver and qpi_deriver != 'unknown-deriver':
            error += 'Deriver `{}` does not exist.  '.format(qpi_deriver)
        if qvd_deriver and qvd_deriver != qpi_deriver:
            error += 'Deriver `{}` does not exist.  '.format(qvd_deriver)
        if error:
            raise RuntimeError(
                error + "Couldn't find deriver for path `{}`".format(path))
        raise RuntimeError(
            'Cannot determine deriver. Is this really a path into the '
            'nix store?', path)

    def add_path(self, path):
        """Add the closure of all derivations referenced by a store path."""
        if not p.exists(path):
            raise RuntimeError('path `{}` does not exist - cannot load '
                               'derivations referenced from it'.format(path))
        _log.debug('Loading derivations referenced by "%s"', path)
        deriver = self._find_deriver(path)
        if self.requisites:
            for candidate in call(['nix-store', '-qR', deriver]).splitlines():
                self.update(candidate)
        else:
            self.update(deriver)

    def update(self, drv_path):
        if not drv_path.endswith('.drv'):
            return
        try:
            drv_obj = load(drv_path)
        except SkipDrv:
            return
        self.derivations.add(drv_obj)

    def load_pkgs_json(self, json_fobj):
        for pkg in json.load(json_fobj).values():
            try:
                patches = pkg['patches']
                if 'known_vulnerabilities' in pkg:
                    patches.extend(pkg['known_vulnerabilities'])
                self.derivations.add(Derive(
                    name=pkg['name'], patches=' '.join(patches)))
            except SkipDrv:
                _log.debug("skipping: {}", pkg)
                continue



from persistent import Persistent
from .utils import compare_versions, haskeys
import functools
import re


@functools.total_ordering
class Vulnerability(Persistent):
    """Represents a single vulnerability.

    Note that this vulnerability can be attached to different products
    and versions, this is called 'configurations' in CPE lingo. We
    represent theses as `Node` objects.
    """

    cve_id = None
    nodes = None
    cvssv3 = 0.0
    cvssv2 = 0.0

    def __init__(self, cve_id, nodes=None, cvssv3=0.0, cvssv2=0.0):
        self.cve_id = cve_id
        self.nodes = nodes or []
        self.cvssv3 = float(cvssv3)
        self.cvssv2 = float(cvssv2)

    def __str__(self):
        return self.cve_id

    def __repr__(self):
        return '<Vulnerability {} (cvssv3: {})>'.format(
            self.cve_id, self.cvssv3)

    def __eq__(self, other):
        return self.cve_id == other.cve_id

    def __lt__(self, other):
        """Order CVEs numerically not lexicographically."""
        (_, s_year, s_id) = self.cve_id.split('-')
        (_, o_year, o_id) = other.cve_id.split('-')
        return (int(s_year), int(s_id)) < (int(o_year), int(o_id))

    def __hash__(self):
        return hash(self.cve_id)

    @classmethod
    def parse(cls, item):
        """Create Vulnerability object from NVD JSON snippet."""
        res = cls(item['cve']['CVE_data_meta']['ID'])
        if 'configurations' in item:
            res.nodes = Node.parse(item['configurations'].get('nodes', []))
        if haskeys(item, 'impact', 'baseMetricV3', 'cvssV3', 'baseScore'):
            res.cvssv3 = float(
                item['impact']['baseMetricV3']['cvssV3']['baseScore'])
        if haskeys(item, 'impact', 'baseMetricV2', 'cvssV2', 'baseScore'):
            res.cvssv2 = float(
                item['impact']['baseMetricV2']['cvssV2']['baseScore'])
        return res

    def match(self, pname, pvers):
        """Returns True if package version is covered by any node."""
        for n in self.nodes:
            if not n.product == pname:
                continue
            if isinstance(n.version, str) and n.version == pvers:
                return True
            elif (isinstance(n.version, list) and
                    all(in_range(pvers, v) for v in n.version)):
                return True
        return False


class Node:
    """CPE configuration item.

    Node uses a simplified matching logic:

    - Don't try to grok the AND/OR logic. NVD data seems to be illogical
      or incomplete in many cases. We accept false positives instead.

    - Currently, vendor strings are ignored completely while matching.
      This may change in a future version.
    """

    vendor = None
    product = None
    version = None

    def __init__(self, vendor, product, version=None):
        self.vendor = vendor
        self.product = product
        self.version = version or []

    @classmethod
    def parse(cls, nodes):
        res = []
        for node in nodes:
            res += cls.parse_matches(node.get('cpe_match', []))
            res += cls.parse(node.get('children', []))
        return res

    R_UNQUOTE = re.compile(r'(\\:)+')

    @classmethod
    def parse_matches(cls, cpe_match):
        nodes = []
        for expr in cpe_match:
            if expr.get('vulnerable') is not True:
                continue
            if 'cpe23Uri' not in expr:
                continue
            cpe23Uri = cls.R_UNQUOTE.sub('-', expr['cpe23Uri'])
            (cpe, cpevers, typ, vendor, product, vers, rev, _) = \
                cpe23Uri.split(':', 7)
            if cpe != 'cpe' or cpevers != '2.3' or typ != 'a':
                continue
            e = cls(vendor, product)
            if vers and vers != '*' and vers != '-':
                if rev and rev != '*' and rev != '-':
                    vers = vers + '-' + rev
                # Exact match: Change self.version to a string with a single
                # version. Doing this, future attempts to apppend() will fail.
                e.version = str(vers)
            else:
                if 'versionStartIncluding' in expr:
                    e.version.append('>=' + expr['versionStartIncluding'])
                if 'versionStartExcluding' in expr:
                    e.version.append('>' + expr['versionStartExcluding'])
                if 'versionEndIncluding' in expr:
                    e.version.append('<=' + expr['versionEndIncluding'])
                if 'versionEndExcluding' in expr:
                    e.version.append('<' + expr['versionEndExcluding'])
            if e.version:
                # no point adding an expr without any version match
                nodes.append(e)
        # dedup strings
        if len(nodes) > 1:
            first = nodes[0]
            for n in nodes[1:]:
                if n.vendor == first.vendor:
                    n.vendor = first.vendor
                if n.product == first.product:
                    n.product = first.product
        return nodes

    def __eq__(self, other):
        return (self.vendor == other.vendor and
                self.product == other.product and
                isinstance(self.version, type(other.version)) and
                self.version == other.version)

    def __repr__(self):
        return '<Node {}:{}, {}>'.format(
            self.vendor, self.product, self.version)


def in_range(pvers, spec):
    """Returns True if package version `pvers` complies with `spec`."""
    if spec[1] == '=':
        if pvers == spec[2:]:
            return True
        idx = 2
    else:
        idx = 1
    if spec[0] == '>':
        return compare_versions(pvers, spec[idx:]) == 1
    elif spec[0] == '<':
        return compare_versions(pvers, spec[idx:]) == -1
    raise ValueError('Invalid version spec', spec)



import collections
import datetime
import logging
import re
import toml
import urllib.parse
import yaml

from vulnix.derivation import split_name
from vulnix.output import Filtered

_log = logging.getLogger(__name__)


# brackets must be followed/preceded immediately by quotation marks
RE_INV_SECT_START = re.compile(r'^\s*\[[^"a-zA-Z]', re.MULTILINE)
RE_INV_SECT_END = re.compile(r'^\s*\[[^\]]*[^"a-zA-Z0-9]\]$', re.MULTILINE)


def check_section_header(content):
    m_begin = RE_INV_SECT_START.search(content)
    m_end = RE_INV_SECT_END.search(content)
    if m_begin or m_end:
        raise RuntimeError(
            'section header must start with \'["\' and end with \'"]\'',
            m_begin, m_end)


def read_toml(content):
    check_section_header(content)
    for k, v in toml.loads(content, collections.OrderedDict).items():
        if len(v.values()) and isinstance(list(v.values())[0], dict):
            raise RuntimeError('malformed section header -- forgot quotes?', k)
        pname, version = split_name(k)
        yield WhitelistRule(pname=pname, version=version, **v)


def read_yaml(content):
    for item in yaml.safe_load(content):
        pname = item.pop('name', None)
        yield WhitelistRule(pname=pname, **item)


def dump_multivalued(val):
    if len(val) == 1:
        return list(val)[0]
    val_l = list(val)
    val_l.sort()
    return val_l


class WhitelistRule:
    """Single whitelist entry.

    Supported fields:
    - pname: package name or `*` for any package
    - version: package version (only if pname is set)
    - cve: affected by CVEs (multi-valued, set)
    - until: this entry will be disabled after the given date (YYYY-MM-DD)
    - issue_url: bug/case ID URLs (multi-valued, set)
    - comment: free form text (multi-valued, list)
    - status (ignored for compatibility reasons)

    The version field may be empty which means that all versions of the
    given package are affected. The package may be "*" to indicate any
    package. In this case, there must be at least one CVE ID present.

    If there are both package and CVE IDs set, only vulnerable items
    which match both are whitelisted.
    """

    pname = None
    version = None

    def __init__(self, **kw):
        for field in ['pname', 'version']:
            self.__dict__[field] = kw.pop(field, None) or '*'
        for field in ['cve', 'issue_url']:
            v = kw.pop(field, [])
            if isinstance(v, set):
                self.__dict__[field] = v
            elif isinstance(v, list):
                self.__dict__[field] = set(v)
            else:
                self.__dict__[field] = set([v])
        if self.pname == '*' and not self.cve:
            raise RuntimeError('either pname or CVE must be set', kw)
        for url in self.issue_url:
            scheme, netloc, path = urllib.parse.urlparse(url)[0:3]
            if not scheme or not netloc or not path:
                raise ValueError('issue must be a valid URL', url)
        v = kw.pop('comment', [])
        self.comment = v if isinstance(v, list) else [v]
        self.until = None
        if 'until' in kw:
            if isinstance(kw['until'], (datetime.datetime, datetime.date)):
                self.until = kw.pop('until')
            else:
                self.until = datetime.datetime.strptime(
                    kw.pop('until'), '%Y-%m-%d').date()
        kw.pop('status', '')  # compat
        if kw:
            _log.warning('Unrecognized whitelist keys: {}'.format(kw.keys()))

    @property
    def name(self):
        if self.version == '*':
            return self.pname
        return '{}-{}'.format(self.pname, self.version)

    def dump(self):
        """Returns this entry as a ready-to-serialize dict.

        If this entry has expired, None is returned. The date in the
        'until' field is already formatted.
        """
        if self.until and self.until <= datetime.date.today():
            return None
        res = collections.OrderedDict()
        for field in ['cve', 'comment', 'issue_url']:
            val = getattr(self, field)
            if val:
                res[field] = dump_multivalued(val)
        if self.until:
            res['until'] = str(self.until)
        return res

    def update(self, other):
        if self.pname != other.pname or self.version != other.version:
            raise RuntimeError(
                'cannot merge rules for different packages', self, other)
        self.cve.update(other.cve)
        if other.until:
            if not self.until or (self.until and other.until > self.until):
                self.until = other.until
        self.issue_url.update(other.issue_url)
        self.comment.extend(other.comment)

    def covers(self, deriv, vulns=set()):
        """Is the given derivation covered by this whitelist item?

        If so, a tuple (match type, whitelist item) is returned.
        """
        if self.pname != '*' and self.pname != deriv.pname:
            return False
        if self.version != '*' and self.version != deriv.version:
            return False
        if self.cve and vulns and \
                self.cve & set(v.cve_id for v in vulns) == set():
            return False
        if self.until and self.until <= datetime.date.today():
            return False
        return True


class Whitelist:
    """Collection of WhitelistRules.

    It can be populated either by reading a config file (`load`) or by
    synthesizing rules from derivations (`add_from`). A whitelist is
    able to save itself into a config file with stable sort order.
    """

    def __init__(self):
        self.entries = collections.OrderedDict()

    def __len__(self):
        return len(self.entries)

    def __getitem__(self, key):
        return self.entries[key]

    def __str__(self):
        """Formats whitelist as string. Used to dump it to a file."""
        return toml.dumps(self.dump()).replace(',]\n', ' ]\n')

    TOML_SECTION_START = re.compile(r'^\[.*\]', re.MULTILINE)
    YAML_SECTION_START = re.compile(r'^-', re.MULTILINE)
    SECTION_FORMAT = re.compile(r'^[a-zA-Z0-9_.*-]+$')

    @classmethod
    def load(cls, fobj):
        """Loads whitelist from file-like object.

        The format (TOML or YAML) is guessed using a heuristic.
        """
        content = fobj.read()
        if isinstance(content, bytes):
            content = content.decode('utf-8')
        filename = ''
        if hasattr(fobj, 'name') and fobj.name:
            filename = fobj.name
        elif hasattr(fobj, 'geturl'):
            filename = fobj.geturl()
        try:
            return cls._parse_cfg(content, filename)
        except (toml.TomlDecodeError, IndexError) as e:
            raise RuntimeError('failed to load `{}`: {}'.format(filename, e))

    @classmethod
    def _parse_cfg(cls, content, filename):
        if filename.endswith('.toml'):
            gen = read_toml(content)
        elif filename.endswith('.yaml'):
            gen = read_yaml(content)
        elif cls.TOML_SECTION_START.search(content):
            gen = read_toml(content)
        elif cls.YAML_SECTION_START.search(content):
            gen = read_yaml(content)
        else:
            raise RuntimeError('cannot detect whitelist format')

        self = cls()
        for rule in gen:
            if not self.SECTION_FORMAT.match(rule.pname):
                raise RuntimeError('invalid package selector', rule.pname)
            self.insert(rule)
        return self

    def dump(self):
        """Serializes whitelist into dict."""
        res = collections.OrderedDict()
        for k, v in self.entries.items():
            _log.debug('Creating WL entry for %s', k)
            entry = v.dump()
            if entry is not None:
                res[k] = entry
        return res

    def candidates(self, pname, version):
        """Matching rules in order of decreasing specificity."""
        try:
            yield self.entries['{}-{}'.format(pname, version)]
        except KeyError:
            pass
        try:
            yield self.entries[pname]
        except KeyError:
            pass
        try:
            yield self.entries['*']
        except KeyError:
            pass

    def find(self, derivation, vulns):
        """Compiles all matching rules into a `Filtered` object."""
        f = Filtered(derivation, vulns)
        for cand in self.candidates(derivation.pname, derivation.version):
            if cand.covers(derivation, vulns):
                f.add(cand)
        return f

    def filter(self, affected):
        return [self.find(deriv, vulns) for (deriv, vulns) in affected.items()]

    def insert(self, rule):
        self.entries[rule.name] = rule

    def update(self, rule):
        name = rule.name
        if name in self.entries:
            self.entries[name].update(rule)
        else:
            self.entries[name] = rule

    def merge(self, other):
        for rule in other.entries.values():
            self.update(rule)

    def add_from(self, filtered_item):
        self.update(WhitelistRule(
            pname=filtered_item.derivation.pname,
            version=filtered_item.derivation.version,
            cve={i.cve_id for i in filtered_item.report}))



from vulnix.vulnerability import Vulnerability
from vulnix.derivation import Derive
from vulnix.output import Filtered, output, output_text, output_json
from vulnix.whitelist import WhitelistRule
import datetime
import json
import pytest

V = Vulnerability


@pytest.fixture
def deriv():
    d = Derive(name='test-0.2')
    d.store_path = '/nix/store/zsawgflc1fq77ijjzb1369zi6kxnc36j-test-0.2'
    return (d, {
        V('CVE-2018-0001'),
        V('CVE-2018-0002'),
        V('CVE-2018-0003', cvssv3=9.8),
    })


@pytest.fixture
def deriv1():
    return (Derive(name='foo-1'), {V('CVE-2018-0004'), V('CVE-2018-0005')})


@pytest.fixture
def deriv2():
    return (Derive(name='bar-2'), {V('CVE-2018-0006', cvssv3=5.0)})


@pytest.fixture
def filt(deriv):
    return Filtered(*deriv)


@pytest.fixture
def items(deriv, deriv1, deriv2):
    return [Filtered(*deriv), Filtered(*deriv1), Filtered(*deriv2)]


def test_init(deriv):
    f = Filtered(*deriv)
    assert f.report == {
        V('CVE-2018-0001'),
        V('CVE-2018-0002'),
        V('CVE-2018-0003')}
    assert not f.masked


def test_add_unspecific_rule(deriv):
    f = Filtered(*deriv)
    f.add(WhitelistRule(pname='test', version='1.2'))
    assert not f.report


def test_add_rule_with_cves(filt):
    filt.add(WhitelistRule(pname='test', version='1.2', cve={'CVE-2018-0001'}))
    assert filt.report == {V('CVE-2018-0002'), V('CVE-2018-0003')}
    assert filt.masked == {V('CVE-2018-0001')}


def test_add_temporary_whitelist(filt):
    assert not filt.until
    filt.add(WhitelistRule(pname='test', version='1.2', until='2018-03-05'))
    assert filt.until == datetime.date(2018, 3, 5)


@pytest.fixture
def wl_items(items):
    # makes deriv1 list only one CVE
    items[1].add(WhitelistRule(
        cve={'CVE-2018-0004'}, issue_url='https://tracker/4'))
    # makes deriv2 disappear completely
    items[2].add(WhitelistRule(pname='bar', comment='irrelevant'))
    return items


def test_output_text(wl_items, capsys):
    output_text(wl_items, show_whitelisted=True)
    assert capsys.readouterr().out == """\
2 derivations with active advisories

------------------------------------------------------------------------
foo-1

CVE                                                CVSSv3
https://nvd.nist.gov/vuln/detail/CVE-2018-0005
https://nvd.nist.gov/vuln/detail/CVE-2018-0004  [whitelisted]

Issue(s):
https://tracker/4

------------------------------------------------------------------------
test-0.2

/nix/store/zsawgflc1fq77ijjzb1369zi6kxnc36j-test-0.2
CVE                                                CVSSv3
https://nvd.nist.gov/vuln/detail/CVE-2018-0003     9.8
https://nvd.nist.gov/vuln/detail/CVE-2018-0001
https://nvd.nist.gov/vuln/detail/CVE-2018-0002

------------------------------------------------------------------------
bar-2

CVE                                                CVSSv3
https://nvd.nist.gov/vuln/detail/CVE-2018-0006     5.0  [whitelisted]

Comment:
* irrelevant
"""


def test_output_json(wl_items, capsys):
    output_json(wl_items)
    assert json.loads(capsys.readouterr().out) == [
        {'affected_by': ['CVE-2018-0005'],
         'derivation': None,
         'name': 'foo-1',
         'pname': 'foo',
         'version': '1',
         'whitelisted': ['CVE-2018-0004'],
         'cvssv3_basescore': {}},
        {'affected_by': ['CVE-2018-0001', 'CVE-2018-0002', 'CVE-2018-0003'],
         'derivation': '/nix/store/zsawgflc1fq77ijjzb1369zi6kxnc36j-test-0.2',
         'name': 'test-0.2',
         'pname': 'test',
         'version': '0.2',
         'whitelisted': [],
         'cvssv3_basescore': {
             'CVE-2018-0003': 9.8,
         }},
    ]


def test_exitcode(items, capsys):
    assert output([], json=True) == 0
    # something to report
    assert output(items) == 2
    # everything masked
    for i in items:
        i.add(WhitelistRule(pname=i.derivation.pname))
    assert output(items) == 0
    assert output(items, show_whitelisted=True) == 1
    capsys.readouterr()  # swallow stdout/stderr: it doesn't matter here



import itertools
import logging
import subprocess
import sys
import tempfile
import time

_log = logging.getLogger(__name__)


def call(cmd):
    """Executes `cmd` and swallow stderr iff returncode is 0."""
    with tempfile.TemporaryFile(prefix='stderr') as capture:
        try:
            output = subprocess.check_output(cmd, stderr=capture)
        except subprocess.CalledProcessError:
            capture.seek(0)
            sys.stderr.write(capture.read().decode('ascii', errors='replace'))
            raise
    return output.decode()


class Timer:

    def __init__(self, section):
        self.section = section

    def __enter__(self):
        _log.debug('>>> %s', self.section)
        self.start = time.clock_gettime(time.CLOCK_MONOTONIC)
        return self

    def __exit__(self, *exc):
        self.end = time.clock_gettime(time.CLOCK_MONOTONIC)
        self.interval = self.end - self.start
        _log.debug('<<< %s %.2fs', self.section, self.interval)
        return False  # re-raise


def components_lt(left, right):
    """Port from nix/src/libexpr/names.cc"""
    try:
        lnum = int(left)
    except (ValueError):
        lnum = None
    try:
        rnum = int(right)
    except (ValueError):
        rnum = None
    if lnum is not None and rnum is not None:
        return lnum < rnum
    if left == '' and rnum is not None:
        return True
    if left == 'pre' and right != 'pre':
        return True
    if right == 'pre':
        return False
    if rnum is not None:
        return True
    if lnum is not None:
        return False
    return left < right


def category(char):
    """Classify `char` into: punctuation, digit, non-digit."""
    if char in ('.', '-'):
        return 0
    if char in ('0', '1', '2', '3', '4', '5', '6', '7', '8', '9'):
        return 1
    return 2


def split_components(v):
    """Yield cohesive groups of digits or non-digits. Skip punctuation."""
    start = 0
    stop = len(v)
    while start < stop:
        cat0 = category(v[start])
        i = start + 1
        while i < stop and category(v[i]) == cat0:
            i += 1
        if cat0 != 0:
            yield v[start:i]
        start = i


def compare_versions(left, right):
    """Compare two versions with the same logic as `nix-env -u`.

    Returns -1 if `left` is older than `right`, 1 if `left` is newer
    than `right`, and 0 if both versions are considered equal.

    See https://nixos.org/nix/manual/#ssec-version-comparisons for rules
    and examples.
    """
    if left == right:
        return 0
    for (lc, rc) in itertools.zip_longest(
            split_components(left), split_components(right), fillvalue=''):
        if lc == rc:
            continue
        if components_lt(lc, rc):
            return -1
        if components_lt(rc, lc):
            return 1
    return 0


def haskeys(d, *keys):
    """Returns True if all keys are present in a nested dict `d`."""
    if len(keys) == 1:
        return keys[0] in d
    first = keys[0]
    if first in d and isinstance(d[first], dict):
        return haskeys(d[first], *keys[1:])
    return False



from vulnix.nix import Store
from vulnix.derivation import Derive
import pkg_resources
import pytest


@pytest.fixture
def json():
    return pkg_resources.resource_stream(
        'vulnix', 'tests/fixtures/pkgs.json')


def test_load_json(json):
    s = Store(requisites=False)
    s.load_pkgs_json(json)
    assert s.derivations == set([
        Derive(name="acpitool-0.5.1", patches="ac.patch battery.patch"),
        Derive(name="aespipe-2.4f"),
        Derive(name="boolector-3.0.0",
               patches="CVE-2019-7560.patch CVE-2019-7559")
    ])



from vulnix.vulnerability import Vulnerability, Node
import json
import pkg_resources

V = Vulnerability


def load(cve):
    return json.loads(pkg_resources.resource_string(
        'vulnix', 'tests/fixtures/{}.json'.format(cve)))


def test_parse_single_matches():
    v = V.parse(load('CVE-2016-2183'))
    # redhat entries do not get compressed, python & openssl do
    assert v.nodes == [
        Node('redhat', 'jboss_enterprise_application_platform', '6.0.0'),
        Node('redhat', 'jboss_enterprise_web_server', '1.0.0'),
        Node('redhat', 'jboss_enterprise_web_server', '2.0.0'),
        Node('redhat', 'jboss_web_server', '3.0'),
        Node('python', 'python', '3.3'),
        Node('python', 'python', '3.4.0'),
        Node('python', 'python', '3.5'),
        Node('python', 'python', '3.6'),
        Node('openssl', 'openssl', '1.0.1a'),
        Node('openssl', 'openssl', '1.0.1b'),
        Node('openssl', 'openssl', '1.0.1c'),
        Node('openssl', 'openssl', '1.0.2a'),
        Node('openssl', 'openssl', '1.0.2b'),
        Node('openssl', 'openssl', '1.0.2c'),
    ]
    assert v.match('openssl', '1.0.1a')
    assert not v.match('openssl', '1.0.1d')


def test_ignore_AND_operator():
    # The AND operators are usually incomplete or incorrect. We prefer to get a
    # few more false positives by ignoring them altogether.
    v = V.parse(load('CVE-2010-0748'))
    assert v.nodes == [
        Node('transmissionbt', 'transmission', ['<1.92'])
    ]
    assert v.match('transmission', '1.91')
    assert not v.match('transmission', '1.92')


def test_parse_escaped_colons():
    cve = {
        'cve': {'CVE_data_meta': {'ID': 'CVE-2014-0477'}},
        'configurations': {'nodes': [{
            'operator': 'OR',
            'cpe_match': [{
                'vulnerable': True,
                'cpe23Uri': 'cpe:2.3:a:email\\:\\:address_module_project:' +
                            'email\\:\\:address:1.902:*:*:*:*:perl:*:*',
            }]
        }]}
    }
    v = V.parse(cve)
    assert len(v.nodes) == 1
    assert v.nodes[0].vendor == 'email-address_module_project'
    assert v.nodes[0].product == 'email-address'


def test_product_not_found(nvd):
    assert [] == nvd.by_product('nonexistent-product')


def test_str():
    assert 'CVE-2016-2183', V(
        'CVE-2016-2183', Node('openssl', 'openssl', '1.0.1a'))


def test_match_multiple_version_ranges():
    v = V.parse(load('CVE-2019-11041'))
    assert v.match('php', '7.2.20')
    assert not v.match('php', '7.2.21')


def test_match_right_version():
    # This CVE contains a non-matching range for openssl and a matching range
    # for an Oracle product. Make sure that the matching algo considers the
    # product name.
    v = V.parse(load('CVE-2015-1789'))
    assert v.match('openssl', '1.0.2-beta1')
    assert not v.match('openssl', '1.1.1d')


def test_compress_strings():
    v = V.parse(load('CVE-2015-1789'))
    assert v.nodes[0].vendor == v.nodes[1].vendor
    assert id(v.nodes[0].vendor) == id(v.nodes[1].vendor)
    assert v.nodes[0].product == v.nodes[1].product
    assert id(v.nodes[0].product) == id(v.nodes[1].product)


def test_cvss_score():
    v = V.parse(load('CVE-2019-10160'))
    assert v.cvssv3 == 9.8
    assert v.cvssv2 == 5.0


def test_no_cvss_score():
    v = V.parse(load('CVE-2010-0748'))
    assert v.cvssv3 == 0.0
    assert v.cvssv2 == 0.0


def test_cve_ordering():
    assert V('CVE-2019-0001') < V('CVE-2019-0002')
    assert V('CVE-2019-0002') == V('CVE-2019-0002')
    assert V('CVE-2019-0001') > V('CVE-2018-0001')
    assert V('CVE-2019-10000') > V('CVE-2018-2000')
    assert V('CVE-2019-0700') > V('CVE-2018-0699')



"""Usage: vulnix {--system | PATH [...]}

vulnix is a tool that scan the NixOS store for packages with known
security issues. There are three main modes of operation:


* Is my NixOS system installation affected?

Invoke:  vulnix --system


* Is my project affected?

Invoke after nix-build:  vulnix ./result


See vulnix --help for a full list of options.
"""


from .nix import Store
from .nvd import NVD, DEFAULT_MIRROR, DEFAULT_CACHE_DIR
from .resource import open_resources
from .utils import Timer
from .whitelist import Whitelist
from .output import output
import click
import logging
import pkg_resources
import sys

CURRENT_SYSTEM = '/nix/var/nix/gcroots/current-system'

_log = logging.getLogger(__name__)


def howto():
    head, tail = __doc__.split('\n', 1)
    click.secho(head, fg='yellow')
    click.echo(tail, nl=False)


def init_logging(verbose):
    logging.getLogger('requests').setLevel(logging.ERROR)
    if verbose >= 2:
        logging.basicConfig(level=logging.DEBUG)
    elif verbose >= 1:
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)


def populate_store(store, gc_roots, profiles, paths, requisites=True):
    """Load derivations from nix store depending on cmdline invocation."""
    if gc_roots:
        store.add_gc_roots()
    for profile in profiles:
        store.add_profile(profile)
    for path in paths:
        store.add_path(path)
    return store


def run(nvd, store):
    """Returns a dict with affected derivations and vulnerabilities."""
    affected = {}
    for derivation in store.derivations:
        vulns = derivation.check(nvd)
        if vulns:
            affected[derivation] = vulns
    _log.debug("Unfiltered affected: %r", affected)
    return affected


@click.command('vulnix')
# what to scan
@click.option('-S', '--system', is_flag=True,
              help='Scan the current system.')
@click.option('-G', '--gc-roots', is_flag=True,
              help='Scan all active GC roots (including old ones).')
@click.option('-p', '--profile', type=click.Path(exists=True),
              multiple=True, help='Scan this profile (eg: ~/.nix-profile)')
@click.option('-f', '--from-file', type=click.File(mode='r'),
              help='Read derivations from file')
@click.argument('path', nargs=-1, type=click.Path(exists=True))
# modify operation
@click.option('-w', '--whitelist', multiple=True, callback=open_resources,
              help='Load whitelist from file or URL (may be given multiple '
              'times).')
@click.option('-W', '--write-whitelist', type=click.File(mode='a'),
              help='Write TOML whitelist containing current matches.')
@click.option('-c', '--cache-dir', type=click.Path(file_okay=False),
              default=DEFAULT_CACHE_DIR,
              help='Cache directory to store parsed archive data. '
              'Default: {}'.format(DEFAULT_CACHE_DIR))
@click.option('-r/-R', '--requisites/--no-requisites', default=True,
              help='Yes: determine transitive closure. No: examine just the '
              'passed derivations (default: yes).')
@click.option('-m', '--mirror',
              help='Mirror to fetch NVD archives from. Default: {}.'.format(
                  DEFAULT_MIRROR),
              default=DEFAULT_MIRROR)
# output control
@click.option('-j', '--json/--no-json', help='JSON vs. human readable output.')
@click.option('-s', '--show-whitelisted', is_flag=True,
              help='Shows whitelisted items as well')
@click.option('-v', '--verbose', count=True,
              help='Increase output verbosity (up to 2 times).')
@click.option('-V', '--version', is_flag=True,
              help='Print vulnix version and exit.')
@click.option('--default-whitelist/--no-default-whitelist', default=True,
              help='(obsolete; kept for compatibility reasons)')
@click.option('-F', '--notfixed', is_flag=True,
              help='(obsolete; kept for compatibility reasons)')
def main(verbose, gc_roots, system, from_file, profile, path, mirror,
         cache_dir, requisites, whitelist, write_whitelist, version, json,
         show_whitelisted, default_whitelist, notfixed):
    if version:
        print('vulnix ' + pkg_resources.get_distribution('vulnix').version)
        sys.exit(0)

    if not (gc_roots or system or profile or path or from_file):
        howto()
        sys.exit(3)

    init_logging(verbose)

    paths = list(path)
    if system:
        paths.append(CURRENT_SYSTEM)

    try:
        with Timer('Load whitelists'):
            wh_sources = whitelist
            whitelist = Whitelist()
            for wl in wh_sources:
                whitelist.merge(Whitelist.load(wl))
        with Timer('Load derivations'):
            store = Store(requisites)
            if from_file:
                if from_file.name.endswith('.json'):
                    _log.debug("loading packages.json")
                    store.load_pkgs_json(from_file)
                else:
                    for drv in from_file.readlines():
                        paths.append(drv.strip())
            else:
                populate_store(store, gc_roots, profile,  paths, requisites)
        with NVD(mirror, cache_dir) as nvd:
            with Timer('Update NVD data'):
                nvd.update()
            with Timer('Scan vulnerabilities'):
                filtered_items = whitelist.filter(run(nvd, store))

            rc = output(filtered_items, json, show_whitelisted)
            if write_whitelist:
                for i in filtered_items:
                    whitelist.add_from(i)
                write_whitelist.close()
                with open(write_whitelist.name, 'w') as f:
                    f.write(str(whitelist))
        sys.exit(rc)

    # This needs to happen outside the NVD context: otherwise ZODB will abort
    # the transaction and we will keep updating over and over.
    except RuntimeError as e:
        _log.exception(e)
        sys.exit(2)
